"""
This is for distributed TensorFlow support.
For an overview of the terminology, the concepts and the technology, see
`here <https://github.com/rwth-i6/returnn/wiki/Distributed-TensorFlow>`_.

Distributed TensorFlow covers multiple levels of functionality:

* Low level:
  TF server and client (TF session) connecting to the server via gRPC.
  ClusterSpec describes the collection of all servers.

* High level via strategies (:class:`tf.distribute.Strategy`).
  See the `official guide <https://www.tensorflow.org/guide/distributed_training>`_.
  This can be used by the Keras API, by the Estimator API,
  and by a custom training loop.

* Concepts and terminology.

RETURNN potentially could support all of this, although it will try to be very explicit about it.
Currently we do not use the high level strategy concept but only the low level functionality.
This implementation was originally discussed `here <https://github.com/rwth-i6/returnn/issues/296>`_.

This is also related to `Horovod <https://github.com/horovod/horovod>`_.
Horovod and distributed TF are orthogonal to each other.
They can both be mixed, or used independently.

This is also related to the dataset pipeline :mod:`TFDataPipeline`.
"""

from __future__ import print_function
import os
import typing
import contextlib
import tensorflow as tf
from tensorflow.python.training.server_lib import ClusterSpec
from tensorflow.python.distribute.cluster_resolver.cluster_resolver import format_master_url
# noinspection PyProtectedMember
from tensorflow.python.distribute.distribute_lib import _DefaultDistributionExtended as DefaultDistributionExtended
from returnn.log import log
from returnn.config import Config
from returnn.util.basic import CollectionReadCheckCovered


class MPIClusterResolver(tf.distribute.cluster_resolver.ClusterResolver):
  """
  ClusterResolver for MPI.
  Distributed TF is in general totally independent of MPI.
  We only use MPI here to figure out the ClusterSpec.
  After this is set up, MPI will not be used anymore.
  TF itself will not make use of MPI;
  all communications are handled via gRPC.
  (Although `Horovod <https://github.com/horovod/horovod>`_ would use MPI, but that is another topic.)

  If you use Sun Grid Engine (SGE) / Oracle Grid Engine,
  with the parallel environment (PE) feature
  (`doc <http://gridscheduler.sourceforge.net/htmlman/htmlman5/sge_pe.html>`_)
  (i.e. the SGE job was started e.g. via: `qsub -pe mpi 8`),
  then you might run your sub processes (slots) via `mpirun`, e.g. like::

    mpirun -np 8 -mca pml ob1 -mca btl ^openib python returnn/rnn.py ...

  Open MPI knows about SGE and will correctly start subprocesses (for each PE slot)
  (potentially remotely).
  From the `Open MPI doc <https://www.open-mpi.org/faq/?category=sge>`_:

    Open MPI will automatically detect when it is running inside SGE and will just "do the Right Thing."
    Specifically, if you execute an mpirun command in a SGE job,
    it will automatically use the SGE mechanisms to launch and kill processes.
    There is no need to specify what nodes to run on -
    Open MPI will obtain this information directly from SGE
    and default to a number of processes equal to the slot count specified.

  SGE provides the `PE_HOSTFILE` env var points to a file which lists all hosts
  and number of slots per host.
  This would be available for the SGE main job process, i.e. where `mpirun` is run,
  but this might *not* be available on the subprocesses (slots)
  which are started by `mpirun` remotely.

  Within such a MPI process, the only reliable way to get information about the other peer processes,
  we must use MPI functions.
  A straight-forward simple way for this is the `mpi4py <https://mpi4py.readthedocs.io/en/stable/>`_ module.
  `mpi4py` can be mixed together with `Horovod <https://github.com/horovod/horovod>`_, so this is a sensible choice.

  This is somewhat similar to :class:`SlurmClusterResolver`.
  Also related: `MPIClusterResolver PR <https://github.com/tensorflow/tensorflow/issues/38356>`_.
  https://github.com/Peidong-Wang/Distributed-TensorFlow-Using-MPI/
  https://stackoverflow.com/questions/10912793/how-are-mpi-processes-started
  """

  def __init__(self):
    # noinspection PyPackageRequirements,PyUnresolvedReferences
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    self._host = MPI.Get_processor_name()
    # Currently just pick some random open port.
    # We could change this to some deterministic variant which could depend on
    # the pid, GPU-number, job-id, or so.
    # But the current variant might actually be fine.
    self._port = _get_open_port()
    self._rank = comm.Get_rank()
    self._size = comm.Get_size()
    hosts = comm.allgather((self._rank, self._host, self._port))
    peers = {rank: "%s:%i" % (host, port) for (rank, host, port) in hosts}
    assert len(peers) == self._size and 0 in peers and (self._size - 1) in peers
    self._peers = [peers[i] for i in range(self._size)]
    self.task_type = "worker"  # currently we are all equal workers (except id 0, which is the chief)
    self.task_id = self._rank
    self.rpc_layer = "grpc"

  def cluster_spec(self):
    """Retrieve the current state of the cluster and return a ClusterSpec.

    Returns:
      A ClusterSpec representing the state of the cluster at the moment this
      function is called.

    Implementors of this function must take care in ensuring that the
    ClusterSpec returned is up-to-date at the time of calling this function.
    This usually means retrieving the information from the underlying cluster
    management system every time this function is invoked and reconstructing
    a cluster_spec, rather than attempting to cache anything.
    """
    # We currently only use the worker task type.
    # We currently do not support in dynamic changes of the cluster environment.
    return ClusterSpec({"worker": self._peers})

  def master(self, task_type=None, task_id=None, rpc_layer=None):
    """Retrieves the name or URL of the session master.

    Args:
      task_type: (Optional) The type of the TensorFlow task of the master.
      task_id: (Optional) The index of the TensorFlow task of the master.
      rpc_layer: (Optional) The RPC protocol for the given cluster.

    Returns:
      The name or URL of the session master.

    Implementors of this function must take care in ensuring that the master
    returned is up-to-date at the time to calling this function. This usually
    means retrieving the master every time this function is invoked.
    """
    task_type = task_type if task_type is not None else self.task_type
    task_id = task_id if task_id is not None else self.task_id

    if task_type is not None and task_id is not None:
      return format_master_url(
        self.cluster_spec().task_address(task_type, task_id),
        rpc_layer or self.rpc_layer)

    return ''

  def num_accelerators(self, task_type=None, task_id=None, config_proto=None):
    """Returns the number of accelerator cores per worker.

    This returns the number of accelerator cores (such as GPUs and TPUs)
    available per worker.

    Optionally, we allow callers to specify the task_type, and task_id, for
    if they want to target a specific TensorFlow process to query
    the number of accelerators. This is to support heterogenous environments,
    where the number of accelerators cores per host is different.

    Args:
      task_type: (Optional) The type of the TensorFlow task of the machine we
        want to query.
      task_id: (Optional) The index of the TensorFlow task of the machine we
        want to query.
      config_proto: (Optional) Configuration for starting a new session to
        query how many accelerator cores it has.

    Returns:
      A map of accelerator types to number of cores.
    """
    # The default might be fine.
    # It will connect to the remote TF server and query that information.
    return super(MPIClusterResolver, self).num_accelerators(
      task_type=task_type, task_id=task_id, config_proto=config_proto)


def _get_open_port():
  """
  https://stackoverflow.com/questions/2838244/get-open-tcp-port-in-python

  :rtype: int
  """
  import socket
  s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
  s.bind(("", 0))
  s.listen(1)
  port = s.getsockname()[1]
  s.close()
  return port


class LocalOnlyClusterResolver(tf.distribute.cluster_resolver.ClusterResolver):
  """
  Cluster resolver for one local instance.
  """

  def __init__(self):
    self._port = _get_open_port()
    self._host = "localhost:%i" % self._port
    self.task_type = "worker"
    self.task_id = 0
    self.rpc_layer = "grpc"

  def cluster_spec(self):
    """
    :rtype: ClusterSpec
    """
    return ClusterSpec({"worker": [self._host]})

  def master(self, task_type=None, task_id=None, rpc_layer=None):
    """Retrieves the name or URL of the session master.

    Args:
      task_type: (Optional) The type of the TensorFlow task of the master.
      task_id: (Optional) The index of the TensorFlow task of the master.
      rpc_layer: (Optional) The RPC protocol for the given cluster.

    Returns:
      The name or URL of the session master.

    Implementors of this function must take care in ensuring that the master
    returned is up-to-date at the time to calling this function. This usually
    means retrieving the master every time this function is invoked.
    """
    task_type = task_type if task_type is not None else self.task_type
    task_id = task_id if task_id is not None else self.task_id

    if task_type is not None and task_id is not None:
      return format_master_url(
        self.cluster_spec().task_address(task_type, task_id),
        rpc_layer or self.rpc_layer)

    return ''


_controller = None  # type: typing.Optional[_Controller]


class _Controller:
  """
  The controller encapsulates the logic needed for distributed TF in RETURNN.
  It would be setup via :func:`init_distributed_tf`.

  We currently check for the TF_CONFIG env var,
  and if set, use the :class:`TFConfigClusterResolver`.
  Otherwise we assume a MPI setup, and use
  :class:`MPIClusterResolver`.

  If we are using in-graph replication, and this is not the chief,
  this would just run a TF server and wait for commands.
  """

  def __init__(self, config):
    """
    :param Config config:
    """
    print("Initialize distributed TensorFlow", file=log.v2)
    self.config = config
    opts = config.get_of_type("distributed_tf", dict, {})
    opts = CollectionReadCheckCovered(opts)
    self.opts = opts
    if opts.get("local_only", False):  # might be useful for testing
      cluster_resolver = LocalOnlyClusterResolver()
      print("Use local-only cluster resolver,", file=log.v4, end=" ")
    elif os.environ.get("TF_CONFIG", ""):
      cluster_resolver = tf.distribute.cluster_resolver.TFConfigClusterResolver()
      print("Use TF_CONFIG %s," % os.environ["TF_CONFIG"], file=log.v4, end=" ")
    else:
      cluster_resolver = MPIClusterResolver()
      print("Use MPI cluster resolver,", file=log.v4, end=" ")
    print("cluster spec %s, master %s" % (cluster_resolver.cluster_spec(), cluster_resolver.master()), file=log.v4)
    self.cluster_resolver = cluster_resolver
    cluster_spec = cluster_resolver.cluster_spec()
    self.cluster_spec = cluster_spec
    tf_session_opts = config.typed_dict.get("tf_session_opts", {})
    server_config = tf.compat.v1.ConfigProto(**tf_session_opts)
    # Note that there is no clean way currently in TF to uninit the TF server.
    # If we would use this multiple times (e.g. in tests),
    # it might actually be better to cache the server as a singleton...
    server = tf.distribute.Server(
      cluster_spec,
      job_name=cluster_resolver.task_type, task_index=cluster_resolver.task_id,
      config=server_config)
    self.server = server
    self.strategy = ReturnnDefaultStrategy()  # not really used currently...
    self.opts.assert_all_read()

  def is_chief(self):
    """
    :return: whether we are the chief (worker 0)
    :rtype: bool
    """
    if "chief" in self.cluster_spec.jobs:
      return self.cluster_resolver.task_type == "chief"
    if "master" in self.cluster_spec.jobs:
      return self.cluster_resolver.task_type == "master"
    if "worker" in self.cluster_spec.jobs:
      if self.cluster_resolver.task_type != "worker":
        return False
      return self.cluster_resolver.task_id == 0
    raise NotImplementedError("is_chief unknown for cluster spec %r" % self.cluster_spec)


# not really used currently
class ReturnnDefaultStrategy(tf.distribute.Strategy):
  """
  RETURNN default strategy.
  """

  def __init__(self):
    super(ReturnnDefaultStrategy, self).__init__(
      extended=ReturnnDefaultStrategyExtended(self))


# not really used currently
# noinspection PyAbstractClass
class ReturnnDefaultStrategyExtended(DefaultDistributionExtended):
  """
  RETURNN default strategy extended.
  """


def init_distributed_tf(config):
  """
  This is called early in startup of RETURNN.

  :param Config config:
  """
  global _controller
  assert not _controller, "init_distributed_tf called twice?"
  _controller = _Controller(config=config)


def is_enabled():
  """
  :rtype: bool
  """
  return bool(_controller)


def get_session_target():
  """
  This would be called if you have a local custom graph in the current process (replica)
  and want to execute parts of it. This is e.g. the case for between-graph replication.
  After creating the graph, you would create a session
  which connects to the server returned by this function.

  :return: URL of the TF server, where the local session should connect to
  :rtype: str
  """
  assert _controller, "init_distributed_tf not called?"
  return _controller.server.target


@contextlib.contextmanager
def _temporary_init_distributed_tf(config):
  """
  This is useful for tests.

  :param config:
  :return: scope where we have initialized distributed TF, and going out-of-scope will uninit again
  """
  global _controller
  init_distributed_tf(config=config)
  yield
  # Now uninit again.
  # Note that there is no real way to stop the server.
  # If we use this often in tests, it might actually be better to cache the server as a singleton...
  assert _controller
  _controller = None  # type: typing.Optional[_Controller]



import theano
import theano.gradient
import theano.tensor as T
import theano.printing
import theano.gof
from theano.sandbox.cuda.basic_ops import (as_cuda_ndarray_variable,
                                           gpu_contiguous)
from theano.gof.opt import OpSub
from theano.compile import optdb
import os

class LSTMOpGrad(theano.sandbox.cuda.GpuOp):
  def __init__(self, inplace):
    self.inplace = inplace
    if inplace:
      #all outputs operate inplace on inputs 4 and 6 (which are DY and H)
      #but when the input is marked multiple times, we get an error
      #so we only mark that output 0 destroys inputs 4 and 6
      #anyway theano knows that inputs 4 and 6 will be destroyed, so it should be OK
      #TODO
      self.destroy_map = {0: [4], 1: [6]}

  def __eq__(self, other):
    return type(self) == type(other) and self.inplace == other.inplace

  def __str__(self):
    if self.inplace:
      return '%s{inplace}' % self.__class__.__name__
    else:
      return '%s{no_inplace}' % self.__class__.__name__

  def __hash__(self):
    return hash(type(self)) ^ hash(self.inplace)

  def make_node(self, V_h, c, idx, Dd, DY, Y, H):
    V_h = gpu_contiguous(as_cuda_ndarray_variable(V_h))
    c = gpu_contiguous(as_cuda_ndarray_variable(c))
    DY = gpu_contiguous(as_cuda_ndarray_variable(DY))
    idx = gpu_contiguous(as_cuda_ndarray_variable(T.cast(idx,'float32')))
    Dd = gpu_contiguous(as_cuda_ndarray_variable(Dd))
    assert V_h.dtype == "float32"
    assert DY.dtype == 'float32'
    assert Y.dtype == 'float32'
    assert H.dtype == 'float32'
    assert c.dtype == 'float32'
    assert V_h.ndim == 2
    assert DY.ndim == 3
    assert Y.ndim == 3
    assert H.ndim == 3
    assert c.ndim == 2
    assert idx.ndim == 2

    return theano.Apply(self, [V_h, c, idx, Dd, DY, Y, H], [H.type(), V_h.type(), c.type()])

  def infer_shape(self, node, input_shapes):
    V_hs, cs, idxs, Dds, DYs, Ys, Hs = input_shapes
    return [Hs, V_hs, cs]

  def c_support_code(self):
    crnn_path = os.path.dirname(__file__)
    with open(crnn_path + "/c_support_code_mdlstm.cpp") as f:
      return f.read()

  def c_code(self, node, name, input_names, output_names, sub):
    V_h, c, i, Dd, DY, Y, H = input_names
    DZ, DV_h, Dc = output_names
    fail = sub['fail']
    inplace = "true" if self.inplace else "false"
    return """

    // std::cout << "LSTMOpGrad called" << std::endl;
    if(!%(inplace)s)
    {
      //std::cout << "warning, inplace optimization failed, not working inplace" << std::endl;
    }

    if(%(DZ)s || %(DV_h)s || %(Dc)s)
    {
      //printf("output storage already exists\\n");
      //TODO check if we can reuse it
      Py_XDECREF(%(DZ)s);
      Py_XDECREF(%(DV_h)s);
      Py_XDECREF(%(Dc)s);
    }

    CudaNdarray * epsilon = 0;
    CudaNdarray * delta = 0;
    if(%(inplace)s)
    {
      epsilon = %(DY)s;
      delta = %(H)s;
      Py_XINCREF(delta);
    }
    else
    {
      epsilon = (CudaNdarray *) CudaNdarray_Copy(%(DY)s);
      delta = (CudaNdarray *) CudaNdarray_Copy(%(H)s);
    }

    const int * H_dim = CudaNdarray_HOST_DIMS(%(H)s);

    int y = 0;
    for(int x = H_dim[0]-1; x >= 0; --x)
    {
      //add recurrent
      bool rightBorder = (x == H_dim[0]-1);
      if(!rightBorder)
      {
        affine_y_x(y, x+1, delta, y, x, %(V_h)s, y, x, epsilon, false, true);
      }

      do_lstm_bwd(delta, epsilon, %(Y)s, %(Dd)s, %(c)s, y, x, rightBorder, %(i)s);

    }

    %(DV_h)s = CudaNdarray_uninitialized_like(%(V_h)s);
    //DV_h = Y[0..end-1]^T * delta[1..end]
    affine_global(%(Y)s, delta, %(DV_h)s, true, false, 1, 0.0f);

    %(DZ)s = delta;

    %(Dc)s = CudaNdarray_uninitialized_like(%(c)s);
    const int * Y_dim = CudaNdarray_HOST_DIMS(%(Y)s);
    cudaMemcpy(CudaNdarray_DEV_DATA(%(Dc)s), CudaNdarray_DEV_DATA(epsilon),
      Y_dim[1]*Y_dim[2]*sizeof(float), cudaMemcpyDeviceToDevice);

    if(!%(inplace)s)
    {
      Py_XDECREF(epsilon);
    }

    """ % locals()

  #!!! change this when changing the code!
  def c_code_cache_version(self):
    return 1, 5

LSTMOpGradNoInplaceInstance = LSTMOpGrad(inplace=False)
LSTMOpGradInplaceInstance = LSTMOpGrad(inplace=True)

LSTMOpGradInplaceOpt = OpSub(LSTMOpGradNoInplaceInstance, LSTMOpGradInplaceInstance)

#hack to avoid being called twice
if not hasattr(optdb, 'LSTMOpGradInplaceOpt_registered'):
  optdb.register('LSTMOpGradInplaceOpt', theano.gof.TopoOptimizer(LSTMOpGradInplaceOpt),
                 50.0, 'fast_run', 'inplace', 'gpuarray')
  optdb.LSTMOpGradInplaceOpt_registered = True


#------------------------

class LSTMOp(theano.sandbox.cuda.GpuOp):
  def __init__(self, inplace):
    self.inplace = inplace
    if inplace:
      #all outputs operate inplace on input 0 (which is Z)
      #but when the input is marked multiple times, we get an error
      #so we only mark that output 0 destroys input 0
      #anyway theano knows that input 0 will be destroyed, so it should be OK
      #TODO
      self.destroy_map = {0: [0]}

  def __eq__(self, other):
    return type(self) == type(other) and self.inplace == other.inplace

  def __str__(self):
    if self.inplace:
      return '%s{inplace}' % self.__class__.__name__
    else:
      return '%s{no_inplace}' % self.__class__.__name__

  def __hash__(self):
    return hash(type(self)) ^ hash(self.inplace)

  def make_node(self, Z, V_h, c, i):
    """
    :param Z: {input,output,forget} gate + cell state. 3d (time,batch,dim*4)
    :param V_h: recurrent matrix. 2d (dim,dim*4)
    :param c: initial cell state. 2d (batch,dim)
    :param i: index. 2d (time,batch) -> 0 or 1
    """
    Z = gpu_contiguous(as_cuda_ndarray_variable(Z))
    V_h = gpu_contiguous(as_cuda_ndarray_variable(V_h))
    c = gpu_contiguous(as_cuda_ndarray_variable(c))
    i = gpu_contiguous(as_cuda_ndarray_variable(T.cast(i,'float32')))
    assert Z.dtype == "float32"
    assert V_h.dtype == "float32"
    assert c.dtype == 'float32'
    assert c.ndim == 2
    assert Z.ndim == 3
    assert i.ndim == 2
    assert V_h.ndim == 2

    # results: output Y, (gates and cell state) H, (final cell state) d
    return theano.Apply(self, [Z, V_h, c, i], [Z.type(), Z.type(), c.type()])

  def c_support_code(self):
    crnn_path = os.path.dirname(__file__)
    with open(crnn_path + "/c_support_code_mdlstm.cpp") as f:
      return f.read()

  def c_code(self, node, name, input_names, output_names, sub):
    Z, V_h, c, i = input_names
    Y, H, d = output_names
    fail = sub['fail']
    inplace = "true" if self.inplace else "false"
    return """
    if(%(Y)s || %(H)s || %(d)s)
    {
      //printf("Y or H or d already exist\\n");
      //TODO check if we can reuse it
      Py_XDECREF(%(Y)s);
      Py_XDECREF(%(H)s);
      Py_XDECREF(%(d)s);
    }

    const int * Z_dim = CudaNdarray_HOST_DIMS(%(Z)s);
    const int dims_Y[] = {Z_dim[0], Z_dim[1], Z_dim[2] / 4};
    const int dims_H[] = {Z_dim[0], Z_dim[1], Z_dim[2]};
    const int dims_d[] = {Z_dim[1], Z_dim[2] / 4};
    int size_d = Z_dim[1] * Z_dim[2] / 4;

    %(Y)s = (CudaNdarray*) CudaNdarray_NewDims(3,dims_Y);
    %(d)s = (CudaNdarray*) CudaNdarray_NewDims(2, dims_d);
    if(%(inplace)s)
    {
      %(H)s = %(Z)s;
      Py_INCREF(%(Z)s);
    }
    else
    {
      %(H)s = (CudaNdarray*) CudaNdarray_NewDims(3,dims_H);
      cudaMemcpy(CudaNdarray_DEV_DATA(%(H)s), CudaNdarray_DEV_DATA(%(Z)s),
      dims_H[0]*dims_H[1]*dims_H[2]*sizeof(float), cudaMemcpyDeviceToDevice);
    }

    int y = 0;
    for(int x = 0; x < Z_dim[0]; ++x)
    {
      if(x > 0)
      {
        //H += Y[x-1]*V_h
        affine_y_x(y, x-1, %(Y)s, y, x, %(V_h)s, y, x, %(H)s);
      }
      float * d_ptr = (x == Z_dim[0] - 1) ? CudaNdarray_DEV_DATA(%(d)s) : 0;
      do_lstm(%(H)s, %(Y)s, %(c)s, d_ptr, y, x, %(i)s);
    }
    """ % locals()

  def grad(self, inputs, output_grads):
    Z, V_h, c, i = inputs
    DY, DH, Dd = output_grads

    Z_raw = Z.owner.inputs[0].owner.inputs[0]
    #TODO!!!
    V_h_raw = V_h.owner.inputs[0]
    c_raw = c.owner.inputs[0].owner.inputs[0]
    i_raw = i.owner.inputs[0].owner.inputs[0]
    #we have to make sure that this in only computed once!
    #for this we have to extract the raw variables before conversion to continuous gpu array
    #so that theano can merge the nodes
    Y, H, d = LSTMOpInstance(Z_raw, V_h_raw, c_raw, i_raw)
    if isinstance(DY.type, theano.gradient.DisconnectedType):
      DY = T.zeros_like(Z)
    if isinstance(Dd.type, theano.gradient.DisconnectedType):
      Dd = T.zeros_like(c)
    DZ, DV_h, Dc = LSTMOpGradNoInplaceInstance(V_h, c, i, Dd, DY, Y, H)
    Di = theano.gradient.grad_undefined(self, 3, inputs[3], 'cannot diff w.r.t. index')

    return [DZ, DV_h, Dc, Di]

  def infer_shape(self, node, input_shapes):
    Zs, V_hs, cs, idxs = input_shapes
    Y_shape = (Zs[0], Zs[1], Zs[2] // 4)
    H_shape = (Zs[0], Zs[1], Zs[2])
    d_shape = (Zs[1], Zs[2] // 4)
    return [Y_shape, H_shape, d_shape]

  #!!! change this when changing the code!
  def c_code_cache_version(self):
    return 1, 6

LSTMOpInstance = LSTMOp(inplace=False)
LSTMOpInplaceInstance = LSTMOp(inplace=True)

LSTMOpInplaceOpt = OpSub(LSTMOpInstance, LSTMOpInplaceInstance)

#hack to avoid begin called twice
if not hasattr(optdb, 'LSTMOpInplaceOpt_registered'):
  optdb.register('LSTMOpInplaceOpt', theano.gof.TopoOptimizer(LSTMOpInplaceOpt),
                 50.0, 'fast_run', 'inplace', 'gpuarray')
  optdb.LSTMOpInplaceOpt_registered = True

class LSTMSOp(theano.sandbox.cuda.GpuOp):
  def __init__(self, inplace):
    self.inplace = inplace
    if inplace:
      #all outputs operate inplace on input 0 (which is Z)
      #but when the input is marked multiple times, we get an error
      #so we only mark that output 0 destroys input 0
      #anyway theano knows that input 0 will be destroyed, so it should be OK
      #TODO
      self.destroy_map = {0: [0]}

  def __eq__(self, other):
    return type(self) == type(other) and self.inplace == other.inplace

  def __str__(self):
    if self.inplace:
      return '%s{inplace}' % self.__class__.__name__
    else:
      return '%s{no_inplace}' % self.__class__.__name__

  def __hash__(self):
    return hash(type(self)) ^ hash(self.inplace)

  def make_node(self, Z, V_h, c, i, att):
    """
    :param Z: {input,output,forget} gate + cell state. 3d (time,batch,dim*4)
    :param V_h: recurrent matrix. 2d (dim,dim*4)
    :param c: initial cell state. 2d (batch,dim)
    :param i: index. 2d (time,batch) -> 0 or 1
    :param att: attention from inverted alignment layer
    """
    Z = gpu_contiguous(as_cuda_ndarray_variable(Z))
    V_h = gpu_contiguous(as_cuda_ndarray_variable(V_h))
    c = gpu_contiguous(as_cuda_ndarray_variable(c))
    i = gpu_contiguous(as_cuda_ndarray_variable(T.cast(i,'float32')))
    att = gpu_contiguous(as_cuda_ndarray_variable(T.cast(att,'float32')))
    assert Z.dtype == "float32"
    assert V_h.dtype == "float32"
    assert c.dtype == 'float32'
    assert c.ndim == 2
    assert Z.ndim == 3
    assert i.ndim == 2
    assert V_h.ndim == 2
    assert att.ndim == 2
    # results: output Y, (gates and cell state) H, (final cell state) d
    return theano.Apply(self, [Z, V_h, c, i, att], [Z.type(), Z.type(), c.type()])

  def c_support_code(self):
    crnn_path = os.path.dirname(__file__)
    with open(crnn_path + "/c_support_code_mdlstm.cpp") as f:
      return f.read()

  def c_code(self, node, name, input_names, output_names, sub):
    Z, V_h, c, i, att = input_names
    Y, H, d = output_names
    fail = sub['fail']
    inplace = "true" if self.inplace else "false"
    return """
    if(%(Y)s || %(H)s || %(d)s)
    {
      //printf("Y or H or d already exist\\n");
      //TODO check if we can reuse it
      Py_XDECREF(%(Y)s);
      Py_XDECREF(%(H)s);
      Py_XDECREF(%(d)s);
    }

    const int * Z_dim = CudaNdarray_HOST_DIMS(%(Z)s);
    const int dims_Y[] = {Z_dim[0], Z_dim[1], Z_dim[2] / 4};
    const int dims_H[] = {Z_dim[0], Z_dim[1], Z_dim[2]};
    const int dims_d[] = {Z_dim[1], Z_dim[2] / 4};
    int size_d = Z_dim[1] * Z_dim[2] / 4;

    %(Y)s = (CudaNdarray*) CudaNdarray_NewDims(3,dims_Y);
    %(d)s = (CudaNdarray*) CudaNdarray_NewDims(2, dims_d);
    if(%(inplace)s)
    {
      %(H)s = %(Z)s;
      Py_INCREF(%(Z)s);
    }
    else
    {
      %(H)s = (CudaNdarray*) CudaNdarray_NewDims(3,dims_H);
      cudaMemcpy(CudaNdarray_DEV_DATA(%(H)s), CudaNdarray_DEV_DATA(%(Z)s),
      dims_H[0]*dims_H[1]*dims_H[2]*sizeof(float), cudaMemcpyDeviceToDevice);
    }

    int y = 0;
    for(int x = 0; x < Z_dim[0]; ++x)
    {
      if(x > 0)
      {
        //H += Y[x-1]*V_h
        affine_y_x(y, x-1, %(Y)s, y, x, %(V_h)s, y, x, %(H)s);
      }
      float * d_ptr = (x == Z_dim[0] - 1) ? CudaNdarray_DEV_DATA(%(d)s) : 0;
      do_lstms(%(H)s, %(Y)s, %(c)s, d_ptr, y, x, %(i)s, %(att)s);
    }
    """ % locals()

  def grad(self, inputs, output_grads):
    Z, V_h, c, i, att = inputs
    DY, DH, Dd = output_grads

    Z_raw = Z.owner.inputs[0].owner.inputs[0]
    #TODO!!!
    V_h_raw = V_h.owner.inputs[0]
    c_raw = c.owner.inputs[0].owner.inputs[0]
    i_raw = i.owner.inputs[0].owner.inputs[0]
    att_raw = att.owner.inputs[0].owner.inputs[0]
    #we have to make sure that this in only computed once!
    #for this we have to extract the raw variables before conversion to continuous gpu array
    #so that theano can merge the nodes
    Y, H, d = LSTMSOpInstance(Z_raw, V_h_raw, c_raw, i_raw, att_raw)
    if isinstance(DY.type, theano.gradient.DisconnectedType):
      DY = T.zeros_like(Z)
    if isinstance(Dd.type, theano.gradient.DisconnectedType):
      Dd = T.zeros_like(c)
    DZ, DV_h, Dc = LSTMOpGradNoInplaceInstance(V_h, c, i, Dd, DY, Y, H)
    Di = theano.gradient.grad_undefined(self, 3, inputs[3], 'cannot diff w.r.t. index')
    Datt = theano.gradient.grad_undefined(self, 4, inputs[4], 'cannot diff w.r.t. index')
    return [DZ, DV_h, Dc, Di, Datt]

  def infer_shape(self, node, input_shapes):
    Zs, V_hs, cs, idxs, atts = input_shapes
    Y_shape = (Zs[0], Zs[1], Zs[2] / 4)
    H_shape = (Zs[0], Zs[1], Zs[2])
    d_shape = (Zs[1], Zs[2] / 4)
    return [Y_shape, H_shape, d_shape]

  #!!! change this when changing the code!
  def c_code_cache_version(self):
    return 1,7.1

LSTMSOpInstance = LSTMSOp(inplace=False)
LSTMSOpInplaceInstance = LSTMSOp(inplace=True)

LSTMSOpInplaceOpt = OpSub(LSTMSOpInstance, LSTMSOpInplaceInstance)

#hack to avoid begin called twice
if not hasattr(optdb, 'LSTMSOpInplaceOpt_registered'):
  optdb.register('LSTMSOpInplaceOpt', theano.gof.TopoOptimizer(LSTMSOpInplaceOpt),
                 50.0, 'fast_run', 'inplace', 'gpuarray')
  optdb.LSTMSOpInplaceOpt_registered = True



import theano
import theano.tensor as T
import numpy
import os
Tfloat = theano.config.floatX  # @UndefinedVariable

#forward backward training with two states per char, no blank, no skips, no priors
class TwoStateHMMOp(theano.Op):
    def __eq__(self, other):
        return type(self) == type(other)

    def __hash__(self):
        return hash(type(self))

    def __str__(self):
        return self.__class__.__name__

    def make_node(self, x, y, seq_lengths,
                  tdp_loop=T.as_tensor_variable(numpy.cast["float32"](0)),
                  tdp_fwd=T.as_tensor_variable(numpy.cast["float32"](0))):
        x = theano.tensor.as_tensor_variable(x)
        assert x.ndim == 3  # tensor: nframes x nseqs x dim
        y = theano.tensor.as_tensor_variable(y)
        assert y.ndim == 2  # matrix: nseqs x max_labelling_length
        seq_lengths = theano.tensor.as_tensor_variable(seq_lengths)
        assert seq_lengths.ndim == 1  # vector of seqs lengths
        assert seq_lengths.dtype == "int32"
        assert tdp_loop.dtype == "float32"
        assert tdp_fwd.dtype == "float32"
        assert tdp_loop.ndim == 0
        assert tdp_fwd.ndim == 0

        return theano.Apply(self, [x, y, seq_lengths, tdp_loop, tdp_fwd], [T.fvector(), T.ftensor3(), T.fmatrix()])
        # first output: CTC error per sequence
        # second output: Derivative w.r.t. Softmax net input

    def c_support_code(self):
        src = ""
        path = os.path.dirname(os.path.abspath(__file__))
        with open(path + '/c_support_code.cpp', 'r') as f:
            src += f.read()
        with open(path + '/two_state_hmm.cpp', 'r') as f:
            src += f.read()
        return src

    def c_compile_args(self):
        return ['-fopenmp']

    def c_code(self, node, name, inp, out, sub):
        x, y, seq_lengths, tdp_loop, tdp_fwd = inp
        errs, err_sigs, priors = out
        fail = sub['fail']
        return """
            Py_XDECREF(%(errs)s);
            Py_XDECREF(%(err_sigs)s);
            Py_XDECREF(%(priors)s);
            npy_intp dims[] = {PyArray_DIM(%(x)s,1)};
            %(errs)s = (PyArrayObject*) PyArray_Zeros(1, dims, PyArray_DescrFromType(NPY_FLOAT32), 0);
            if (!%(errs)s)
                %(fail)s;
            %(err_sigs)s = (PyArrayObject*) PyArray_Zeros(PyArray_NDIM(%(x)s), PyArray_DIMS(%(x)s), PyArray_DescrFromType(NPY_FLOAT32), 0);
            if (!%(err_sigs)s)
                %(fail)s;
            %(priors)s = (PyArrayObject*) PyArray_Zeros(PyArray_NDIM(%(x)s) - 1, PyArray_DIMS(%(x)s) + 1, PyArray_DescrFromType(NPY_FLOAT32), 0);
            if (!%(priors)s)
                %(fail)s;
            {
                ArrayF errsWr(%(errs)s);
                ArrayF errSigsWr(%(err_sigs)s);
                ArrayF priorsWr(%(priors)s);
                ArrayF xWr(%(x)s);
                ArrayI yWr(%(y)s);
                CArrayI seqLensWr(%(seq_lengths)s);

                CArrayF tdp_loopWr(%(tdp_fwd)s);
                CArrayF tdp_fwdWr(%(tdp_loop)s);
                //convert from log space back to probabilities
                float tdp_loop = exp(tdp_loopWr());
                float tdp_fwd = exp(tdp_fwdWr());

                /*errsWr.debugPrint("errsWr");
                errSigsWr.debugPrint("errSigsWr");
                xWr.debugPrint("xWr");
                yWr.debugPrint("yWr");
                seqLensWr.debugPrint("seqLensWr");*/

                int numSeqs = seqLensWr.dim(0);
                #pragma omp parallel for
                for(int i = 0; i < numSeqs; ++i)
                {
                    TwoStateHMM hmm;
                    SArrayF errSigsSWr(errSigsWr, 1, i);
                    SArrayF priorsSWr(priorsWr, 0, i);
                    hmm.forwardBackward(CSArrayF(xWr, 1, i), CSArrayI(yWr, 0, i), seqLensWr(i), errsWr(i),
                      errSigsSWr, priorsSWr, tdp_loop, tdp_fwd);
                }
            }
        """ % locals()

    #IMPORTANT: change this, if you change the c-code
    def c_code_cache_version(self):
        return (1.6,)




import theano.tensor as T
from returnn.theano.util import complex_bound
import numpy


def relu(z):
  # Use fastest implementation.
  # https://github.com/Theano/Theano/issues/2698
  # https://github.com/Lasagne/Lasagne/pull/163#issuecomment-81806482
  return (z + abs(z)) / 2.0

def clipped01lu(z):
  """
  0 for x <= 0
  x for 0 <= x <= 1
  1 for 1 <= x
  """
  # Not sure about the fastest implementation...
  return relu(z) - relu(z - numpy.float32(1))

def clippedlu(z):
  """
  -1 for  x <= -1
   x for -1 <=  x <= 1
   1 for  1 <=  x
  """
  # Not sure about the fastest implementation...
  return relu(z + numpy.float32(1)) - relu(z - numpy.float32(1)) - numpy.float32(1)

def elu(z): # http://arxiv.org/pdf/1511.07289v1.pdf
  return T.switch(T.ge(z,0), z, T.exp(z) - 1)

def identity(z):
  return z

def softsign(z):
  return z / (1.0 + abs(z))

def softsquare(z):
  return 1 / (1.0 + z * z)

def maxout(z):
  return T.max(z, axis=0)

def softmax(z):
  assert z.ndim >= 1
  if z.ndim <= 2:
    return T.nnet.softmax(z)
  else:
    from returnn.theano.util import time_batch_make_flat
    z_flat = time_batch_make_flat(z)
    assert z_flat.ndim == 2
    return T.reshape(T.nnet.softmax(z_flat), z.shape)

def log_softmax(z):
  assert z.ndim >= 1
  if z.ndim <= 2:
    return T.nnet.logsoftmax(z)
  else:
    from returnn.theano.util import time_batch_make_flat
    z_flat = time_batch_make_flat(z)
    assert z_flat.ndim == 2
    return T.reshape(T.nnet.logsoftmax(z_flat), z.shape)

def gauss(z):
  return T.exp(-T.sqr(z))

def cdf(z):
  """Cumulative distribution function via erf (Error function)"""
  return (numpy.float32(1) + T.erf(z)) / numpy.float32(2)

def constant_one():
  return 1

def constant_zero():
  return 0


# from https://github.com/MatthieuCourbariaux/BinaryNet/blob/master/Train-time/binary_net.py
from theano.scalar.basic import UnaryScalarOp, same_out_nocomplex
from theano.tensor.elemwise import Elemwise
class Round3(UnaryScalarOp):
  def c_code(self, node, name, x, z, sub):
    x, = x
    z, = z
    return "%(z)s = round(%(x)s);" % locals()

  def grad(self, inputs, gout):
    (gz,) = gout
    return gz,


round3_scalar = Round3(same_out_nocomplex, name='round3')
round3 = Elemwise(round3_scalar)


def hard_sigmoid(x):
  return T.clip((x + 1.) / 2., 0, 1)


# The neurons' activations binarization function
# It behaves like the sign function during forward propagation
# And like:
#   hard_tanh(x) = 2*hard_sigmoid(x)-1
# during back propagation
def binary_tanh(x):
  return 2. * round3(hard_sigmoid(x)) - 1.


def binary_sigmoid(x):
  return round3(hard_sigmoid(x))


ActivationFunctions = {
  'logistic': T.nnet.sigmoid,
  'sigmoid': T.nnet.sigmoid,  # alias
  'tanh': T.tanh,
  'relu': relu,
  'clipped01lu': clipped01lu,
  'clippedlu': clippedlu,
  'elu': elu,
  'identity': identity,
  'one': constant_one,
  'zero': constant_zero,
  'softsign': softsign,
  'softsquare': softsquare,
  'maxout': maxout,
  'sin': T.sin,
  'cos': T.cos,
  'complex_bound': complex_bound,
  'softmax': softmax,
  'log_softmax': log_softmax,
  'gauss': gauss,
  "erf": T.erf,
  "exp": T.exp,
  "abs": T.abs_,
  "sqr": T.sqr,
  "sqrt": T.sqrt,
  "binary_sigmoid" : binary_sigmoid,
  "binary_tanh" : binary_tanh,
  "cdf": cdf
}


def strtoact(act):
  """
  :type act: str | list[str]
  :param act: activation function name, or multiple such as a list or separated by ":"
  :rtype: theano.Op | list[theano.Op]
  """
  if isinstance(act, (list, tuple)):
    return [strtoact(a) for a in act]
  if ":" in act:
    return [strtoact(a) for a in act.split(":")]
  assert act in ActivationFunctions, "invalid activation function: %s" % act
  return ActivationFunctions[act]

def strtoact_single_joined(act):
  """
  :type act: str | None
  :param act: activation function name, or multiple such as a list or separated by ":"
  :rtype: theano.Op
  """
  if not act:
    return identity
  if ":" in act:
    joined = identity
    for f in [strtoact_single_joined(a) for a in act.split(":")]:
      joined = lambda x: f(joined(x))
    return joined
  assert act in ActivationFunctions, "invalid activation function: %s" % act
  return ActivationFunctions[act]
